from django.db import models
from django.db.models import Lookup
from django.db.models.fields import Field

from dirtyfields import DirtyFieldsMixin


@Field.register_lookup
class Like(Lookup):
    """Allows doing LIKE comparisons when querying."""

    lookup_name = "like"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s LIKE %s" % (lhs, rhs), params


class BaseModel(DirtyFieldsMixin, models.Model):
    """Common fields for all models."""

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Person(BaseModel):
    """A person."""

    first_name = models.CharField(max_length=256)
    last_name = models.CharField(max_length=256)
    roles = models.ManyToManyField("Role", through="PersonRole", related_name="persons")

    def __str__(self):
        return "{} {} ({})".format(self.first_name, self.last_name, self.pk)

    def __repr__(self):
        return "{}(id={!r}, first_name={!r}, last_name={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.first_name,
            self.last_name,
        )


class Role(BaseModel):
    """A role variant."""

    type = models.SlugField(max_length=64, unique=True)
    name_nb = models.CharField(max_length=256)
    name_en = models.CharField(max_length=256)
    meta = models.JSONField(null=True, blank=True)

    def __str__(self):
        return str(self.name_nb or self.name_en or self.slug)

    def __repr__(self):
        return "{}(id={!r}, type={!r}, name_nb={!r}, name_en={!r}, meta={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.type,
            self.name_nb,
            self.name_en,
            self.meta,
        )


class PersonRole(BaseModel):
    """The relationship between a person and a role."""

    person = models.ForeignKey(
        "Person", on_delete=models.PROTECT, related_name="person_roles"
    )
    role = models.ForeignKey(
        "Role", on_delete=models.PROTECT, related_name="person_roles"
    )

    # class Meta:
    #     constraints = [
    #         models.UniqueConstraint(
    #             fields=["person", "role"], name="personrole_person_role_unique"
    #         )
    #     ]

    def __repr__(self):
        return "{}(id={!r}, person={!r}, role={!r})".format(
            self.__class__.__name__, self.pk, self.person, self.role
        )


class Notification(BaseModel):
    """A change notification that should be delivered to a message queue."""

    identifier = models.IntegerField(null=True, blank=True)
    object_type = models.TextField()
    operation = models.TextField()
    issued_at = models.IntegerField()
    meta = models.JSONField(null=True, blank=True)

    def __repr__(self):
        return "{}(id={!r}, identifier={!r}, object_type={!r}, operation={!r}, issued_at={!r}), meta={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.identifier,
            self.object_type,
            self.operation,
            self.issued_at,
            self.meta,
        )
